from demo import custom_unit_testing as ut
import nprime.pyprime as p      # import from pyprime file
from nprime import is_prime     # import from main package
from nprime.plot import sacks_plot, ulam_plot


def main():
    """ example of the prime function and custom tests """
    # Demo of the Primary testing functions
    print(is_prime(7))
    print(p.fermat(7))
    print(p.miller_rabin(7))

    # Demo of the Prime generating functions
    print(p.generate_primes(7))
    print(p.find_primes(2, 7, p.fermat))

    # Demo of the custom_test functions
    print(ut.custom_test(p.is_prime))

    # Demo of Graphical Prime functions
    sacks_plot()
    ulam_plot()


if __name__ == '__main__':
    main()



import unittest

from nprime.pyprime import postponed_sieve_eratosthenes
from tests.prime_testcase import FIRST_PRIMES
from itertools import islice


class TestPostponedSieve(unittest.TestCase):
    """ Tests for postponed sieve of eratosthenes """

    def test_first_few_primes(self):
        """ Making sure that all keys from the trial_division are prime  """
        generator = postponed_sieve_eratosthenes()
        first_primes = list(islice(generator, 0, len(FIRST_PRIMES)))
        self.assertEqual(FIRST_PRIMES, first_primes)


if __name__ == '__main__':
    unittest.main()



import unittest

from nprime.pyprime import trial_division as trial_division
from tests.prime_testcase import FIRST_PRIMES


class TestTrialDivision(unittest.TestCase):
    """ Tests for generate prime function """

    def test_001_is_not_integer_return_error(self):
        """ there should only be 2 in the list, if limit is set to 2 """
        self.assertRaises(ValueError, trial_division, 3.8)

    def test_002_is_inferior_to_two_return_error(self):
        """ there should only be 2 in the list, if limit is set to 2 """
        self.assertRaises(ValueError, trial_division, -1)

    def test_003_is_trial_division_showing_primes_and_composite(self):
        """ The trial_division should work for the first 10 numbers"""
        self.assertEqual({2: [4, 6, 8], 3: [6, 9], 5: [], 7: []}, trial_division(10))

    def test_004_are_all_keys_primes(self):
        """ Making sure that all keys from the trial_division are prime  """
        primes = list(trial_division(70).keys())
        primes.sort()
        self.assertEqual(FIRST_PRIMES, primes)


if __name__ == '__main__':
    unittest.main()



import unittest

from nprime.pyprime import generate_primes
from tests.prime_testcase import FIRST_PRIMES


class TestGeneratesPrimes(unittest.TestCase):
    """ Tests for generate prime function """

    def test_001_is_there_2_in_list_of_first_prime(self):
        """ there should only be 2 in the list, if limit is set to 2 """
        self.assertEqual([2], generate_primes(2))

    def test_002_inferior_to_two_return_nothing(self):
        """ No prime should be generated if upper is inferior to 2"""
        self.assertEqual([], generate_primes(1))

    def test_003_are_first_prime_all_generated(self):
        output = generate_primes(70)
        for n in range(0, len(output)):
            self.assertEqual(FIRST_PRIMES[n], output[n],
                              msg=f'Missing - {FIRST_PRIMES[n]} - in generated prime list ')


if __name__ == '__main__':
    unittest.main()



import unittest

from nprime.pyprime import pyprime


class TestPyPrime(unittest.TestCase):
    """ test for pyprime """

    def setUp(self):
        """ Set up of the tests """
        pass

    def test_001_is_zero_not_prime(self):
        """Is zero correctly determined not to be prime?"""
        self.assertFalse(pyprime(0))

    def test_002_is_one_not_prime(self):
        """Is four correctly determined not to be prime?"""
        self.assertFalse(pyprime(1), msg='Four is not prime!')

    def test_003_is_three_prime(self):
        """Is two correctly determined to be prime?"""
        self.assertTrue(pyprime(3), msg='Three is prime!')

    def test_004_negative_number_not_prime(self):
        """Is a negative number correctly determined not to be prime?"""
        for index in range(-1, -10, -1):
            self.assertFalse(pyprime(index), msg=f'{index} should not be determined to be prime')

    def test_004_float_return_error_message(self):
        """ An error message is raised for a float """
        self.assertRaises(TypeError, pyprime, 3.8)

    def test_005_bool_return_error_message(self):
        """ An error message is raised for a bool """
        self.assertRaises(TypeError, pyprime, True)

    def test_006_string_return_error_message(self):
        """ An error message is raised for a String"""
        self.assertRaises(TypeError, pyprime, "three")

    def tearDown(self):
        """ tear down of the tests """
        pass


if __name__ == '__main__':
    unittest.main()



import unittest

from nprime.pyprime import is_prime, fermat, miller_rabin
from tests.prime_testcase import make_test_case


class TestIsPrime(make_test_case(is_prime)):
    """ Tests for miller rabin function """
    pass


class TestFermat(make_test_case(fermat)):
    """ Tests for fermat function """
    pass


class TestMillerRabin(make_test_case(miller_rabin)):
    """ Tests for miller rabin function """
    pass


if __name__ == '__main__':
    unittest.main()



import unittest

from nprime.pyprime import find_primes
from tests.prime_testcase import FIRST_PRIMES


class TestFindPrimes(unittest.TestCase):
    """ Tests for findprimes function """

    def test_001_lower_bigger_than_upper_error_message(self):
        """ An error message is raised for a float """
        self.assertRaises(ValueError, find_primes, 100, 50)

    def test_002_lower_is_equal_to_one_error_message(self):
        """ An error message is raised for a float """
        self.assertRaises(ValueError, find_primes, 1, 50)

    def test_003_lower_is_inferior_to_one_error_message(self):
        """ An error message is raised for a float """
        self.assertRaises(ValueError, find_primes, 0, 50)

    def test_004_lower_equal_upper_error_message(self):
        """ An error message is raised for a float """
        self.assertRaises(ValueError, find_primes, 50, 50)

    def test_005_are_first_prime_all_discovered(self):
        """ Test that first primes are found """
        output = find_primes(2, 70)
        for n in range(0, len(output)):
            self.assertEqual(FIRST_PRIMES[n], output[n],
                             msg=f'Missing - {FIRST_PRIMES[n]} - in generated prime list ')

    def test_006_are_primes_found_in_interval(self):
        """ Test that primes are found in a different interval """
        prime_list_1000_to_1100 = [1009, 1013, 1019, 1021, 1031, 1033, 1039, 1049, 1051, 1061, 1063, 1069,
                                   1087, 1091, 1093, 1097]
        self.assertTrue(find_primes(1000, 1100) == prime_list_1000_to_1100)

    def test_007_find_no_primes_when_no_primes(self):
        """ Test that it founds no primes in an interval with no primes """
        self.assertTrue(find_primes(370270, 370370) == [])


if __name__ == '__main__':
    unittest.main()



"""
nprime main application package

Auto generate command:
    mkinit nprime/__init__.py --nomods -w
"""
from nprime.pyprime import (fermat, find_primes, generate_primes, is_prime,
                            miller_rabin, pyprime, sacks, sieve_eratosthenes,
                            trial_division, ulam, )

__submodules__ = [
    'pyprime',
]

__version__ = '1.3.1'

__all__ = ['fermat', 'find_primes', 'generate_primes', 'is_prime',
           'miller_rabin', 'pyprime', 'sacks', 'sieve_eratosthenes',
           'trial_division', 'ulam']



from nprime.pyprime import sacks, pyprime, ulam


def sacks_plot(upper=10000, prime_test_function=pyprime):  # pragma: no cover
    """
    Render the sacks_plot from the sacks function.
    By default the coord is plot in white and the prime_coord in black

    Return a polar plot of the sacks' diagram

    """
    import matplotlib.pyplot as plt
    coord, prime_coord = sacks(upper, prime_test_function)

    plt.figure()
    ax = plt.subplot(111, projection='polar', facecolor='white')
    plt.title('Sacks\' Diagram', loc='right')
    ax.plot(zip(*coord), "w+", markersize=1)
    ax.plot(zip(*prime_coord), "ko", markersize=2)
    plt.show()


def ulam_plot(upper=10000, edge=4, prime_test_function=pyprime):  # pragma: no cover
    """
    Render the sacks_plot from the ulam function.
    By default the coord is plot in white and the prime_coord in black

    Return a polar plot of the ulam's spiral

    """
    import matplotlib.pyplot as plt
    coord, prime_coord = ulam(upper, edge, prime_test_function)

    plt.figure()
    plt.title('Ulam\'s sprial', loc='right')
    plt.plot(zip(*coord), 'w+', markersize=1)
    plt.plot(zip(*prime_coord), 'ko', markersize=2)
    plt.grid(True)
    plt.show()



""" nprime package """
from setuptools import setup, find_packages


def convert(markdown_path):
    """Convert a Markdown file to a reStructuredText file with the pypandoc"""
    try:
        import pypandoc
        output = pypandoc.convert_file(markdown_path, 'rst')
    except(IOError, ImportError, OSError):
        with open(markdown_path, 'r', encoding='utf-8') as f:
            output = f.read()
    return output


LONG_DESCRIPTION = convert("README.md")

setup(name='nprime',
      description='Python library for primes algorithms',
      long_description=LONG_DESCRIPTION,
      author='sylhare',
      author_email='sylhare@outlook.com',
      url='https://github.com/Sylhare/nprime',
      license='GNU General Public License v3.0',
      tests_require=['pytest'],
      install_requires=['matplotlib>=3.5.0'],
      keywords=['prime',
                'fermat',
                'miller rabin',
                'math'],
      packages=find_packages(),
      package_data={
          'License': ['LICENSE'],
          'Readme': ['README.md'],
      },
      platforms='any',
      zip_safe=True,
      test_suite='tests',
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Programming Language :: Python :: 3.12",
          "Programming Language :: Python :: 3.13",
          "Programming Language :: Python :: 3 :: Only",
          "Environment :: Other Environment",
          "Intended Audience :: Developers",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Education",
          "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
          "Operating System :: OS Independent",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: Software Development :: Libraries",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Utilities"]
      )



import math
import random


# -- Prime testing functions -- #
def is_prime(n):
    """
    Check if the number "n" is prime, with n > 1.

    Returns a boolean, True if n is prime.

    Example:
        >>> is_prime(101)
        True
        >>> is_prime(102)
        False
        >>> is_prime(103)
        True
    """
    for i in range(2, math.isqrt(n) + 1):
        if n % i == 0:
            return False
    return True


def fermat(n, t=100):
    """
    Probabilistic algorithm
    Taking "t" randoms "a" and testing the Fermat's theorem on number "n" > 1

    Prime probability is right is 1 - 1/(2^t)
    Returns a boolean: True if n passes the tests.

    Example:
        >>> fermat(101)
        True
        >>> fermat(102)
        False
        >>> fermat(103)
        True
    """
    for _ in range(0, t):
        a = random.randrange(1, n)
        x = pow(a, n - 1, n)  # (a^(n-1)) modulo n

        if x == 1:
            prime = True  # /!\ Probable prime
        else:
            prime = False
            break

    return prime


def miller_rabin(n, t=100):
    """
    A probabilistic algorithm which determines
    whether a given number (n > 1) is prime or not.
    The miller_rabin tests is repeated t times to get more accurate results.

    Returns a boolean: True if n passes the tests

    Example:
        >>> miller_rabin(101)
        True
        >>> miller_rabin(102)
        False
        >>> miller_rabin(103)
        True
    """
    if n == 2:
        prime = True  # To normalize and make the algorythm works with 2
    else:
        prime = False  # All other even number will output false

    # Step 1: Have n-1 = 2^s * m (with m odd, and s number of twos factored)
    d = n - 1
    s = 0
    while d % 2 == 0:
        d //= 2  # d equals to quotient of d divided 2
        s += 1  # s > 1 when n is odd

    for _ in range(0, t):
        #  Step 2: test (a^d)^2^r ≡ 1 mod n for all r
        a = random.randrange(1, n)
        for _ in range(0, s):
            x = pow(a, d * pow(2, s), n)
            if x == 1 or x == -1:
                prime = True  # Should be true for all a
            else:
                return False  # When not true, it's not prime for sure

    return prime  # /!\ Probable prime


# Prime generating functions #
def generate_primes(upper=0):
    """
    Generate a list of primes from 2 to a set limit
    It re-uses the sieve of Eratosthenes, but does not tag composites

    Returns a list of integer.

    Example:
        >>> upper = 50
        >>> primes = list(generate_primes(upper))
        >>> print(primes)
        [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47]
    """

    if upper >= 2:
        primes = [2]

        for n in range(3, upper + 1):
            # We only check if n is divided by the previous primes
            sqrt_n = math.isqrt(n)
            divisor = None
            for p in primes:
                if sqrt_n < p:
                    break
                if n % p == 0:
                    divisor = p
                    break  # not prime
            if divisor is None:
                primes.append(n)  # must be prime
    else:
        primes = []

    return primes


def postponed_sieve_eratosthenes(start=0):
    """
    An infinite generator for primes using a "sliding sieve" of erathostenes
    based on a starting value.

    Algorithm from Will Ness and Tim Peters [SO2211990]_, based on ActiveState
    code from David Eppstein and Alex Martelli [ASR117119]_, with ideas used by
    Guy Blelloch [CMUCACM8]_ and likely others. Adjustment for including a
    starting point adapted from [SO69336435]_.

    References:
        .. [SO2211990] https://stackoverflow.com/questions/2211990/efficient-infinite-generator-prime/10733621#10733621
        .. [ASR117119] https://code.activestate.com/recipes/117119-sieve-of-eratosthenes/
        .. [CMUCACM8] https://www.cs.cmu.edu/%7Escandal/cacm/node8.html
        .. [SO69336435] https://stackoverflow.com/a/69345662/887074

    Args:
        start (int): start generating primes greater than this number.

    Yields:
        int

    Example:
        >>> from nprime.pyprime import *  # NOQA
        >>> from itertools import islice
        >>> generator = postponed_sieve_eratosthenes()
        >>> list(islice(generator, 0, 5))
        [2, 3, 5, 7, 11]
        >>> list(islice(generator, 0, 5))
        [13, 17, 19, 23, 29]

    Example:
        >>> from nprime.pyprime import *  # NOQA
        >>> from itertools import islice
        >>> # Generate larger primes in the billions
        >>> generator = postponed_sieve_eratosthenes(start=500_000_000)
        >>> list(islice(generator, 0, 5))
        [500000003, 500000009, 500000041, 500000057, 500000069]

    Example:
        >>> # xdoctest: +SKIP("can stress older hardware")
        >>> from nprime.pyprime import *  # NOQA
        >>> from itertools import islice
        >>> # Generate large primes in the quadrillions
        >>> generator = postponed_sieve_eratosthenes(start=20_000_000_000_000_000)
        >>> next(generator)
        20000000000000003
    """
    import itertools
    yield from (p for p in (2, 3, 5, 7) if p >= start)
    # holds current multiples of each base prime (i.e. below the sqrt of the
    # current production point), together with their step values
    multiples = {}
    ps = postponed_sieve_eratosthenes()
    next(ps)  # skip 2
    p = next(ps)  # start at p=3
    psq = p * p   # track the square of the prime (starts at 9)

    first_cand = psq

    # Adjust the initial state
    while psq < start:
        s = 2 * p
        # Avoid using math.ceil due to precision loss
        # m = p + s * math.ceil((start - p) / s)   # multiple of p
        if (start - p) % s == 0:
            m = start
        else:
            m = start + s - ((start - p) % s)
        while m in multiples:
            m += s
        multiples[m] = s
        p = next(ps)
        psq = p * p

    if start > first_cand:
        first_cand = start + (1 - start % 2)  # ensure first candidate is odd

    for cand in itertools.count(first_cand, 2):
        if cand in multiples:
            # candidate is composite
            step = multiples.pop(cand)
        elif cand < psq:
            # candiates is prime
            yield cand
            continue
        else:
            # candidate reached the square of the base prime
            # move to next base prime
            step = 2 * p
            p = next(ps)
            psq = p * p
        cand += step
        while cand in multiples:
            cand += step
        multiples[cand] = step


def sieve_eratosthenes(upper):
    """
    Implementation of the sieve of erathostenes that discover the primes and their composite up to a limit.

    :return: a dictionary,
                the key are the primes up to n
                the value is the list of composites of these primes up to n

    Example:
        >>> upper = 10
        >>> sieve_eratosthenes(upper)
        {2: [4, 6, 8, 10], 3: [9], 5: [], 7: []}
    """
    primes = {}
    composites = set()
    for p in range(2, upper + 1):
        if p not in composites:
            primes[p] = list(range(p * p, upper + 1, p))
            composites.update(primes[p])

    return primes


def trial_division(upper):
    """
    Implementation of the trial division that discover the primes and all their divider up to a limit.

    :return: a dictionary,
                the key are the primes up to n
                the value is the list of composites of these primes up to n

    Example:
        >>> upper = 10
        >>> trial_division(upper)
        {2: [4, 6, 8], 3: [6, 9], 5: [], 7: []}
    """

    if isinstance(upper, int) and upper >= 2:
        primes = {2: []}

        for n in range(3, upper):
            prime = True

            for key in primes:
                if n % key == 0:
                    primes[key].append(n)
                    prime = False

            if prime:
                primes[n] = []

    else:
        raise ValueError(" Input must be an integer > 2")

    return primes


def find_primes(lower, upper, prime_test_function=is_prime):
    """
    Find the number of primes between lower and upper range.
    We should have 1 < lower < upper.

    primeTest determines the function used to tests the prime of the number
    primeTest is by default is_prime() and should return a boolean

    Returns a list of prime integer.

    Example:
        >>> upper = 100
        >>> lower = 2
        >>> find_primes(lower, upper)
        [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    """
    if upper <= lower or lower <= 1:
        raise ValueError("We should have 1 < lower < upper")

    primes = []

    for n in range(lower, upper + 1):
        if prime_test_function(n):
            primes.append(n)

    return primes


# Other functions #
def pyprime(n, func=is_prime):
    """
    Function that will encapsulate and tests n before doing function(n)
    It is mostly to encapsulate the primality testing functions
    Basically making sure that:
        n is a number
        n > 1

    Example:
        >>> pyprime(0, is_prime)
        False
        >>> pyprime(29, is_prime)
        True
    """
    # assert will trigger an error if the input is not correct
    if type(n) is int:
        if n > 1:
            return func(n)
        else:
            return False
    else:
        raise TypeError("n should be an integer")


# Graphical Prime functions #
def sacks(upper=1000, prime_test_function=pyprime):  # pragma: no cover
    """
    Generate the sack diagram values up to a set limit (upper)

    primeTest determines the function used to tests the primality of the number
    primeTest is by default is_prime() and should return a boolean

    Returns tho lists:
        1- The none prime polar coordinates: coord
        2- The prime polar coordinates: prime_coord

    Example:
        >>> coord, prime_coord = sacks(100)
        >>> assert len(prime_coord) == 25
        >>> assert len(coord) == 75
    """
    coord = []  # Normal numbers' polar value
    prime_coord = []  # Prime numbers' polar value

    for i in range(0, upper):  # A rotation is made for each perfect square,
        theta = math.sqrt(i) * 2 * math.pi  # i=1 theta= 2pi for a given i, angle=(i*theta)/1
        r = math.sqrt(i)

        if prime_test_function(i):
            prime_coord.append((theta, r))
        else:
            coord.append((theta, r))
    return coord, prime_coord


def ulam(upper=1000, edge=4, prime_test_function=pyprime):  # pragma: no cover
    """
    Ulam's spiral aim to represent the primes and none primes in a spiral way

    edge (edge>3) determines the polygone size by the number of edges, 3 triangle, 4 rectangle, 5 Pentagone ...
    For odd number of edge, the spiral gets misaligned

    primeTest determines the function used to tests the primality of the number
    primeTest is by default is_prime() and should return a boolean

    Returns tho lists:
        1- The none prime polar coordinates: coord
        2- The prime polar coordinates: prime_coord

    Example:
        >>> coord, prime_coord = sacks(100)
        >>> assert len(prime_coord) == 25
        >>> assert len(coord) == 75
    """
    theta = 0  # Keep track of the spiral rotation
    psi = math.radians(360 / edge)  # Angle of the polygone's corner

    turn = 3  # Threshold that indicates to turn at the end of each edge's length
    length = 0  # length of the edge, gets bigger as it spirals
    spiral = 2  # Threshold that indicates when to increase the length of an edge
    spiral_increment = int(edge / 2)  # when the edge length has to go up to spiral

    coord = [(0, 0)]  # Other numbers' coordinates
    prime_coord = []  # Primes' coordinates
    x = 0
    y = 0

    for i in range(2, upper):
        if i == spiral:
            length += 1
            spiral = length * spiral_increment + i

        if i == turn:
            theta += psi
            turn = i + length

        x += math.cos(theta)
        y += math.sin(theta)

        if prime_test_function(i):
            prime_coord.append((x, y))
        else:
            coord.append((x, y))

    return coord, prime_coord



"""
Make prime test case that can be reusable for any prime finding function
"""
import unittest

# First primes that the function should succeed at finding
FIRST_PRIMES = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67]

# First non prime that the function should confirm as not prime
FIRST_NOT_PRIMES = [4, 6, 8, 9, 10, 12, 14, 15, 16, 18, 20, 21, 22, 24, 25, 26, 27, 28]

# Carmichael number often trigger false positive for the fermat algorithm
CARMICHAEL = [561, 1105, 1729, 2465, 2821, 6601, 8911, 10585, 15841, 29341]

# The key is the base, the list is the pseudoprimes of that base
PSEUDO_PRIMES = {2: [2047, 3277, 4033, 4681, 8321],
                 3: [121, 703, 1891, 3281, 8401, 8911],
                 4: [341, 1387, 2047, 3277, 4033, 4371],
                 5: [781, 1541, 5461, 5611, 7813],
                 6: [217, 481, 1111, 1261, 2701],
                 7: [25, 325, 703, 2101, 2353, 4525],
                 8: [9, 65, 481, 511, 1417, 2047],
                 9: [91, 121, 671, 703, 1541, 1729]}


def make_test_case(prime_function):
    """Make prime test case"""

    class PrimeTestCase(unittest.TestCase):
        def test_001_is_two_prime(self):
            """Is two correctly determined to be prime?"""
            self.assertTrue(prime_function(2), msg='Two is prime!')

        def test_002_first_primes_prime(self):
            """ test if true for the first known primes """
            for n in FIRST_PRIMES:
                self.assertTrue(prime_function(n))

        def test_003_is_four_not_prime(self):
            """Is four correctly determined not to be prime?"""
            self.assertFalse(prime_function(4), msg='Four is not prime!')

        def test_004_first_non_primes_not_prime(self):
            """ test false for the first non primes """
            for n in FIRST_NOT_PRIMES:
                self.assertFalse(prime_function(n), msg=f'{n} should not be determined to be prime')

        def test_005_pseudoprimes_numbers_not_prime(self):
            """ test False for pseudo primes """
            for _, value in PSEUDO_PRIMES.items():
                for n in value:
                    self.assertFalse(prime_function(n), msg=f'{n} should not be determined to be prime')

        def test_006_carmichael_numbers_not_prime(self):
            """ Test false for carmichael numbers """
            for n in CARMICHAEL:
                self.assertFalse(prime_function(n), msg=f'{n} should not be determined to be prime')

    return PrimeTestCase


if __name__ == '__main__':
    unittest.main()
